#import "HFImageEditorViewController.h"

static const CGFloat kMaxUIImageSize = 1024;
static const CGFloat kPreviewImageSize = 120;
static const CGFloat kDefaultCropWidth = 320;
static const CGFloat kDefaultCropHeight = 320;
static const CGFloat kBoundingBoxInset = 15;
static const CGFloat kMaxZoomScale = 4;
static const CGFloat kMinZoomScale = 0.5;
static const NSTimeInterval kAnimationIntervalReset = 0.25;
static const NSTimeInterval kAnimationIntervalTransform = 0.2;

@interface HFImageEditorViewController (){
    bool isPanning;
    float lastScale;
    CGPoint panTouch;
    UIView *currentDragView;
    CGPoint previousFrameCenter;
    
    UIView *topView;
    UIView *bottomView;
    UIView *leftView;
    UIView *rightView;
    
    UIView *topLeftView;
    UIView *topRightView;
    UIView *bottomLeftView;
    UIView *bottomRightView;
    
    UIImageView *tlnode;
    UIImageView *trnode;
    UIImageView *blnode;
    UIImageView *brnode;
}
@property (nonatomic,retain) UIImageView *imageView;
@property (nonatomic,assign) CGRect cropRect;
@property (retain, nonatomic) IBOutlet UIPanGestureRecognizer *panRecognizer;
@property (retain, nonatomic) IBOutlet UIRotationGestureRecognizer *rotationRecognizer;
@property (retain, nonatomic) IBOutlet UIPinchGestureRecognizer *pinchRecognizer;
@property (retain, nonatomic) IBOutlet UITapGestureRecognizer *tapRecognizer;
@property (nonatomic,retain) IBOutlet UIView<HFImageEditorFrame> *frameView;
@property (nonatomic, retain) IBOutlet UIToolbar *toolbar;

@property(nonatomic,assign) NSUInteger gestureCount;
@property(nonatomic,assign) CGPoint touchCenter;
@property(nonatomic,assign) CGPoint rotationCenter;
@property(nonatomic,assign) CGPoint scaleCenter;
@property(nonatomic,assign) CGFloat scale;


@end



@implementation HFImageEditorViewController

@synthesize doneCallback = _doneCallback;
@synthesize originalImage = _originalImage;
@synthesize sourceImage = _sourceImage;
@synthesize previewImage = _previewImage;
@synthesize cropSize = _cropSize;
@synthesize outputWidth = _outputWidth;
@synthesize frameView = _frameView;
@synthesize imageView = _imageView;
@synthesize panRecognizer = _panRecognizer;
@synthesize rotationRecognizer = _rotationRecognizer;
@synthesize tapRecognizer = _tapRecognizer;
@synthesize pinchRecognizer = _pinchRecognizer;
@synthesize touchCenter = _touchCenter;
@synthesize rotationCenter = _rotationCenter;
@synthesize scaleCenter = _scaleCenter;
@synthesize scale = _scale;
@synthesize minimumScale = _minimumScale;
@synthesize maximumScale = _maximumScale;
@synthesize gestureCount = _gestureCount;
@synthesize toolbar= _toolbar;

@dynamic panEnabled;
@dynamic rotateEnabled;
@dynamic scaleEnabled;
@dynamic tapToResetEnabled;
@dynamic cropBoundsInSourceImage;


#define IMAGE_CROPPER_OUTSIDE_STILL_TOUCHABLE 40.0f
#define IMAGE_CROPPER_INSIDE_STILL_EDGE 20.0f


#ifndef CGOriginX
#define CGOriginX(rect)                 rect.origin.x
#endif

#ifndef CGOriginY
#define CGOriginY(rect)                 rect.origin.y
#endif

- (void)initialize
{
    _panRecognizer = [[UIPanGestureRecognizer alloc] initWithTarget:self action:@selector(handlePan:)];
    _rotationRecognizer = [[UIRotationGestureRecognizer alloc] initWithTarget:self action:@selector(handleRotation:)];
    _pinchRecognizer = [[UIPinchGestureRecognizer alloc] initWithTarget:self action:@selector(handlePinch:)];
    _tapRecognizer = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(handleTap:)];

}

#pragma mark Properties

- (void)setCropSize:(CGSize)cropSize
{
    _cropSize = cropSize;
    self.cropRect = CGRectMake((self.frameView.bounds.size.width-self.cropSize.width)/2,
                               (self.frameView.bounds.size.height-self.cropSize.height)/2,
                               self.cropSize.width, self.cropSize.height);
    
    [self updateCropRect];
}

- (CGSize)cropSize
{
    if(_cropSize.width == 0 || _cropSize.height == 0) {
        _cropSize = CGSizeMake(kDefaultCropWidth, kDefaultCropHeight);
    }
    return _cropSize;
}

- (UIImage *)previewImage
{
    if(_previewImage == nil && _sourceImage != nil) {
        if(self.sourceImage.size.height > kMaxUIImageSize || self.sourceImage.size.width > kMaxUIImageSize) {
            CGFloat aspect = self.sourceImage.size.height/self.sourceImage.size.width;
            CGSize size;
            if(aspect >= 1.0) { //square or portrait
                size = CGSizeMake(kPreviewImageSize,kPreviewImageSize*aspect);
            } else { // landscape
                size = CGSizeMake(kPreviewImageSize,kPreviewImageSize*aspect);
            }
            _previewImage = [self scaledImage:self.sourceImage  toSize:size withQuality:kCGInterpolationLow];
            
        } else {
            _previewImage = _sourceImage;
        }
    }
    return  _previewImage;
}

- (void)setSourceImage:(UIImage *)sourceImage
{
    if(sourceImage != _sourceImage) {
        _sourceImage = sourceImage;
        self.previewImage = nil;
    }
}


- (void)updateCropRect
{
    int margin = 30;
    float aspectX = (self.frameView.bounds.size.width-(2*margin)) / self.cropRect.size.width;
    float aspectY = (self.frameView.bounds.size.height-(2*margin)) / self.cropRect.size.height;
    float aspect = MIN(aspectX, aspectY);
    
    CGRect temp = self.frameView.cropRect;
    
    
    
    self.cropRect = CGRectMake(((self.frameView.bounds.size.width-(self.cropSize.width * aspect))/2),
                               ((self.frameView.bounds.size.height-(self.cropSize.height * aspect))/2),
                               (self.cropSize.width * aspect),
                               (self.cropSize.height * aspect));
    
    self.frameView.cropRect = self.cropRect;

    
    CGPoint moveVector = CGPointMake((self.frameView.cropRect.origin.x + (self.frameView.cropRect.size.width/2) - (temp.origin.x + (temp.size.width/2))),
                                     (self.frameView.cropRect.origin.y + (self.frameView.cropRect.size.height/2)) - (temp.origin.y + (temp.size.height/2)));

    CGAffineTransform imgTransformMove =  CGAffineTransformMakeTranslation(moveVector.x, moveVector.y);
    CGAffineTransform imgTransformScale = CGAffineTransformMakeScale(aspect, aspect);
    CGAffineTransform transform = CGAffineTransformConcat(imgTransformMove, imgTransformScale);
    CGAffineTransform imgTransform = CGAffineTransformConcat(self.imageView.transform, transform);
    
    [self refreshNodes];
    
    void (^doUpdate)(void) = ^{
        [self.imageView setTransform: imgTransform];
    };
    
    self.view.userInteractionEnabled = NO;
    [UIView animateWithDuration:kAnimationIntervalReset animations:doUpdate completion:^(BOOL finished) {
        self.view.userInteractionEnabled = YES;
    }];

}


- (void)setPanEnabled:(BOOL)panEnabled
{
    self.panRecognizer.enabled = panEnabled;
}

- (BOOL)panEnabled
{
    return self.panRecognizer.enabled;
}

- (void)setScaleEnabled:(BOOL)scaleEnabled
{
    self.pinchRecognizer.enabled = scaleEnabled;
}

- (BOOL)scaleEnabled
{
    return self.pinchRecognizer.enabled;
}


- (void)setRotateEnabled:(BOOL)rotateEnabled
{
    self.rotationRecognizer.enabled = rotateEnabled;
}

- (BOOL)rotateEnabled
{
    return self.rotationRecognizer.enabled;
}

- (void)setTapToResetEnabled:(BOOL)tapToResetEnabled
{
    self.tapRecognizer.enabled = tapToResetEnabled;
}

- (BOOL)tapToResetEnabled
{
    return self.tapToResetEnabled;
}

#pragma mark Public methods
-(void)reset:(BOOL)animated
{
    CGFloat aspect = self.sourceImage.size.height/self.sourceImage.size.width;
    CGFloat w = CGRectGetWidth(self.cropRect);
    CGFloat h = aspect * w;
    self.scale = 1;
    
    void (^doReset)(void) = ^{
        self.imageView.transform = CGAffineTransformIdentity;
        self.imageView.frame = CGRectMake(CGRectGetMidX(self.frameView.frame) - w/2, CGRectGetMidY(self.frameView.frame) - h/2,w,h);
    };
    if(animated) {
        self.view.userInteractionEnabled = NO;
        [UIView animateWithDuration:kAnimationIntervalReset animations:doReset completion:^(BOOL finished) {
            self.view.userInteractionEnabled = YES;
        }];
    } else {
        doReset();
    }
}



-(void)refreshNodes{
    int nodeSize = 50;
    tlnode.frame = CGRectMake(self.cropRect.origin.x - nodeSize/2,
                              self.cropRect.origin.y -nodeSize/2,
                              nodeSize,
                              nodeSize);
    
    trnode.frame = CGRectMake(self.cropRect.origin.x + self.cropRect.size.width - nodeSize/2,
                              self.cropRect.origin.y -nodeSize/2,
                              nodeSize,
                              nodeSize);
    
    blnode.frame = CGRectMake(self.cropRect.origin.x - nodeSize/2,
                              self.cropRect.origin.y + self.cropRect.size.height - nodeSize/2,
                              nodeSize,
                              nodeSize);
    
    brnode.frame = CGRectMake(self.cropRect.origin.x + self.cropRect.size.width - nodeSize/2,
                              self.cropRect.origin.y + self.cropRect.size.height - nodeSize/2,
                              nodeSize,
                              nodeSize);
}


#pragma mark View Lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
    UIImageView *imageView = [[UIImageView alloc] init];
    
    [self.view insertSubview:imageView belowSubview:self.frameView];
    self.imageView = imageView;
    //self.imageView.center = self.frameView.center;
    [self.view setMultipleTouchEnabled:YES];

    [self initialize];
    isPanning = NO;
    self.frameView.bounds = CGRectMake(0,
                                       0,
                                       self.view.frame.size.width,
                                       self.view.frame.size.height - self.toolbar.frame.size.height);
    self.imageView.bounds = self.frameView.bounds;
    
    self.panRecognizer.cancelsTouchesInView = NO;
    self.panRecognizer.delegate = self;
    [self.frameView addGestureRecognizer:self.panRecognizer];
    self.rotationRecognizer.cancelsTouchesInView = NO;
    self.rotationRecognizer.delegate = self;
    [self.frameView addGestureRecognizer:self.rotationRecognizer];
    self.pinchRecognizer.cancelsTouchesInView = NO;
    self.pinchRecognizer.delegate = self;
    [self.frameView addGestureRecognizer:self.pinchRecognizer];
    self.tapRecognizer.numberOfTapsRequired = 2;
    [self.frameView addGestureRecognizer:self.tapRecognizer];
        
    
    UIImage *nodeImage = [UIImage imageNamed:@"node.png"];
    tlnode = [[UIImageView alloc]initWithImage:nodeImage];
    trnode = [[UIImageView alloc]initWithImage:nodeImage];
    blnode = [[UIImageView alloc]initWithImage:nodeImage];
    brnode = [[UIImageView alloc]initWithImage:nodeImage];
    
    [tlnode setContentMode:UIViewContentModeCenter];
    [trnode setContentMode:UIViewContentModeCenter];
    [blnode setContentMode:UIViewContentModeCenter];
    [brnode setContentMode:UIViewContentModeCenter];
    
    tlnode.autoresizingMask = UIViewAutoresizingFlexibleRightMargin | UIViewAutoresizingFlexibleBottomMargin;
    trnode.autoresizingMask = UIViewAutoresizingFlexibleLeftMargin | UIViewAutoresizingFlexibleBottomMargin;
    blnode.autoresizingMask = UIViewAutoresizingFlexibleRightMargin | UIViewAutoresizingFlexibleTopMargin;
    brnode.autoresizingMask = UIViewAutoresizingFlexibleLeftMargin | UIViewAutoresizingFlexibleTopMargin;
    
    [self.frameView addSubview:tlnode];
    [self.frameView addSubview:trnode];
    [self.frameView addSubview:blnode];
    [self.frameView addSubview:brnode];
}


- (void)viewDidUnload
{
    [super viewDidUnload];
    
    [self setFrameView:nil];
    [self setImageView:nil];
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{
    return (interfaceOrientation == UIInterfaceOrientationPortrait);
}

- (void)viewWillAppear:(BOOL)animated
{
    [self.imageView setAlpha:0.0f];
    [super viewWillAppear:animated];
    self.originalImage = self.sourceImage;
    self.previewImage = self.originalImage;
    self.imageView.image = self.previewImage;
    

//    self.cropSize = self.sourceImage.size;
//    [self updateCropRect];
//    [self reset:YES];
//    [self refreshNodes];
}

-(void)viewDidAppear:(BOOL)animated{
    self.cropSize = self.sourceImage.size;
    [self updateCropRect];
    [self reset:YES];
    [self refreshNodes];
    
    
    [self updateCropRect];
    [UIView animateWithDuration:kAnimationIntervalReset delay:0 options:UIViewAnimationOptionCurveEaseIn animations:^{
        [self.imageView setAlpha:1.0f];
    } completion:nil];
}

#pragma mark Actions

-(IBAction)getOriginalImage:(id)sender{
    self.sourceImage = self.originalImage;
    self.previewImage = self.originalImage;
    self.imageView.image = self.previewImage;
    self.cropSize = self.sourceImage.size;
    [self reset:YES];
}

- (IBAction)resetAction:(id)sender
{
    [self reset:NO];
}

- (IBAction)resetAnimatedAction:(id)sender
{
    [self reset:YES];
}


- (IBAction)doneAction:(id)sender
{
    [self cropImage];
}

-(void)cropImage{
    UIActivityIndicatorView *spinner = [[UIActivityIndicatorView alloc]initWithFrame:CGRectMake(0, 0, self.view.frame.size.width, self.view.frame.size.height)];
    [self.view addSubview: spinner];
    [spinner setActivityIndicatorViewStyle:UIActivityIndicatorViewStyleWhiteLarge];
    [spinner setBackgroundColor:[UIColor colorWithWhite:0 alpha:0.8f]];
    [spinner setColor:[UIColor colorWithRed: (255/255.0f) green:(10/255.0f) blue:(10/255.0f) alpha:1]];
    spinner.hidden = NO;
    [spinner setHidesWhenStopped:YES];
    [spinner startAnimating];
    
    self.view.userInteractionEnabled = NO;
    [self startTransformHook];
    
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
        
        CGImageRef resultRef = [self newTransformedImage:self.imageView.transform
                                             sourceImage:self.sourceImage.CGImage
                                              sourceSize:self.sourceImage.size
                                       sourceOrientation:self.sourceImage.imageOrientation
                                             outputWidth:self.outputWidth ? self.outputWidth : self.sourceImage.size.width
                                                cropSize:self.cropRect.size
                                
                                           imageViewSize:self.imageView.bounds.size];
        dispatch_async(dispatch_get_main_queue(), ^{
            UIImage *transformedImg =  [UIImage imageWithCGImage:resultRef scale:1.0 orientation:UIImageOrientationUp];
            self.imageView.image = transformedImg;
            
            self.sourceImage = transformedImg;
            self.previewImage = transformedImg;
            CGImageRelease(resultRef);
            
            
            NSString *notificationName = @"CroppedNotify";
            [[NSNotificationCenter defaultCenter] postNotificationName:notificationName object:nil userInfo:nil];

            [spinner stopAnimating];
            
//            self.view.userInteractionEnabled = YES;
//            if(self.doneCallback){
//                self.doneCallback(transformedImg, NO);
//            }
//            [self endTransformHook];
//            [self reset:YES];
        });
    });

}



- (IBAction)cancelAction:(id)sender
{
    if(self.doneCallback) {
        self.doneCallback(nil, YES);
    }
    [[self navigationController] popViewControllerAnimated:YES];
}

#pragma mark Touches

- (void)handleTouches:(NSSet*)touches
{
    self.touchCenter = CGPointZero;
    if(touches.count < 2){
        isPanning = NO;
        CGFloat insetAmount = IMAGE_CROPPER_INSIDE_STILL_EDGE;
        
        CGPoint touch = [[touches anyObject] locationInView:self.frameView];
        if (CGRectContainsPoint(CGRectInset(self.cropRect, insetAmount, insetAmount), touch)) {
            isPanning = YES;
            panTouch = touch;
            return;
        }
        
        CGRect frame = self.cropRect;
        CGFloat x = touch.x;
        CGFloat y = touch.y;
        
        currentDragView = nil;
        
        if (CGRectContainsPoint(CGRectInset(tlnode.frame, -insetAmount, -insetAmount), touch)) {
            currentDragView = tlnode;
            
            if (CGRectContainsPoint(tlnode.frame, touch)) {
                
                if((tlnode.frame.origin.x + tlnode.frame.size.width < trnode.frame.origin.x
                    && tlnode.frame.origin.y + tlnode.frame.size.height < blnode.frame.origin.y)
                    ||(CGOriginX(frame) - x >= 0
                       && CGOriginY(frame) - y >= 0)
                    ){
                    frame.size.width += CGOriginX(frame) - x;
                    frame.size.height += CGOriginY(frame) - y;
                    frame.origin = touch;
                }                
            }
        }
        else if (CGRectContainsPoint(CGRectInset(trnode.frame, -insetAmount, -insetAmount), touch)) {
            currentDragView = trnode;
            
            if (CGRectContainsPoint(trnode.frame, touch)) {
                
                if((trnode.frame.origin.x - trnode.frame.size.width > tlnode.frame.origin.x
                    && trnode.frame.origin.y + trnode.frame.size.height < brnode.frame.origin.y)
                   ||( CGOriginY(frame) - y >= 0
                      && x - CGOriginX(frame) >= 0)
                   ){
                
                    frame.size.height += CGOriginY(frame) - y;
                    frame.origin.y = y;
                    frame.size.width = x - CGOriginX(frame);
               }
            }
        }
        else if (CGRectContainsPoint(CGRectInset(blnode.frame, -insetAmount, -insetAmount), touch)) {
            currentDragView = blnode;
            
            if (CGRectContainsPoint(blnode.frame, touch)) {
                
                if((blnode.frame.origin.x + blnode.frame.size.width < brnode.frame.origin.x
                   && blnode.frame.origin.y > tlnode.frame.origin.y + tlnode.frame.size.height)
                   ||(CGOriginX(frame) - x >= 0
                      && y - CGOriginY(frame) <= 0)
                   ){
                    
                    frame.size.width += CGOriginX(frame) - x;
                    frame.size.height = y - CGOriginY(frame);
                    frame.origin.x =x;
                }
            }
        }
        else if (CGRectContainsPoint(CGRectInset(brnode.frame, -insetAmount, -insetAmount), touch)) {
            currentDragView = brnode;
            
            if (CGRectContainsPoint(brnode.frame, touch)) {
                
                if((brnode.frame.origin.x > blnode.frame.origin.x + blnode.frame.size.width
                    && brnode.frame.origin.y > trnode.frame.origin.y + trnode.frame.size.height)
                   ||(x - CGOriginX(frame) >= 0
                      && y - CGOriginY(frame) >= 0)
                   ){
                    frame.size.width = x - CGOriginX(frame);
                    frame.size.height = y - CGOriginY(frame);
               }
            }
        }
        
        self.cropRect = frame;
        
        [self updateBounds];
    }
    else{
        [touches enumerateObjectsUsingBlock:^(id obj, BOOL *stop) {
            UITouch *touch = (UITouch*)obj;
            CGPoint touchLocation = [touch locationInView:self.imageView];
            self.touchCenter = CGPointMake(self.touchCenter.x + touchLocation.x, self.touchCenter.y +touchLocation.y);
        }];
        self.touchCenter = CGPointMake(self.touchCenter.x/touches.count, self.touchCenter.y/touches.count);
    }
}

-(void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event {
    [self handleTouches:[event allTouches]];
}

-(void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event {
    [self handleTouches:[event allTouches]];
}

-(void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event {
    if(!isPanning){
        self.cropSize = self.cropRect.size;
    }
}

-(void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event {
   [self handleTouches:[event allTouches]];
}


- (void)updateBounds {

    CGRect frame = self.cropRect;
    CGFloat x = CGOriginX(frame);
    CGFloat y = CGOriginY(frame);
    CGFloat width = frame.size.width;
    CGFloat height = frame.size.height;
    
    
    tlnode.frame = CGRectMake(x - tlnode.frame.size.width/2.0,
                              y - tlnode.frame.size.height/2.0,
                              tlnode.frame.size.width,
                              tlnode.frame.size.height);

    trnode.frame = CGRectMake(x + width -trnode.frame.size.width/2.0,
                              y - trnode.frame.size.height/2.0,
                              trnode.frame.size.width,
                              trnode.frame.size.height);
    
    blnode.frame = CGRectMake(x -blnode.frame.size.width/2.0,
                              y + height - blnode.frame.size.height/2.0,
                              blnode.frame.size.width,
                              blnode.frame.size.height);
    
    brnode.frame = CGRectMake(x + width - brnode.frame.size.width/2.0,
                              y + height - brnode.frame.size.height/2.0,
                              brnode.frame.size.width,
                              brnode.frame.size.height);
    
    self.frameView.cropRect = CGRectMake(tlnode.frame.origin.x + tlnode.frame.size.width/2.0,
                                        tlnode.frame.origin.y + tlnode.frame.size.height/2.0,
                                        brnode.frame.origin.x - blnode.frame.origin.x,
                                        brnode.frame.origin.y - trnode.frame.origin.y);
    
    [self didChangeValueForKey:@"crop"];
}

#pragma mark Gestures

- (BOOL)handleGestureState:(UIGestureRecognizerState)state
{
    BOOL handle = YES;
    switch (state) {
        case UIGestureRecognizerStateBegan:
            self.gestureCount++;
            break;
        case UIGestureRecognizerStateCancelled:
        case UIGestureRecognizerStateEnded: {
            self.gestureCount--;
            handle = NO;
            if(self.gestureCount == 0) {
                CGFloat scale = self.scale;
                if(self.minimumScale != 0 && self.scale < self.minimumScale) {
                    scale = self.minimumScale;
                } else if(self.maximumScale != 0 && self.scale > self.maximumScale) {
                    scale = self.maximumScale;
                }
                if(scale != self.scale) {
                    CGFloat deltaX = self.scaleCenter.x-self.imageView.bounds.size.width/2.0;
                    CGFloat deltaY = self.scaleCenter.y-self.imageView.bounds.size.height/2.0;
                    
                    CGAffineTransform transform =  CGAffineTransformTranslate(self.imageView.transform, deltaX, deltaY);
                    transform = CGAffineTransformScale(transform, scale/self.scale , scale/self.scale);
                    transform = CGAffineTransformTranslate(transform, -deltaX, -deltaY);
                    self.view.userInteractionEnabled = NO;
                    [UIView animateWithDuration:kAnimationIntervalTransform delay:0 options:UIViewAnimationOptionCurveEaseIn animations:^{
                        self.imageView.transform = transform;            
                    } completion:^(BOOL finished) {
                        self.view.userInteractionEnabled = YES;
                        self.scale = scale;
                    }];
                    
                }
            }
        } break;
        default:
            break;
    }
    return handle;
}


- (IBAction)handlePan:(UIPanGestureRecognizer*)recognizer
{
        if([self handleGestureState:recognizer.state] && isPanning) {
        CGPoint translation = [recognizer translationInView:self.imageView];
        CGAffineTransform transform = CGAffineTransformTranslate( self.imageView.transform, translation.x, translation.y);
        self.imageView.transform = transform;

        [recognizer setTranslation:CGPointMake(0, 0) inView:self.frameView];
    }

}

- (IBAction)handleRotation:(UIRotationGestureRecognizer*)recognizer
{
    if([self handleGestureState:recognizer.state]) {
        if(recognizer.state == UIGestureRecognizerStateBegan){
            self.rotationCenter = self.touchCenter;
        } 
        CGFloat deltaX = self.rotationCenter.x-self.imageView.bounds.size.width/2;
        CGFloat deltaY = self.rotationCenter.y-self.imageView.bounds.size.height/2;

        CGAffineTransform transform =  CGAffineTransformTranslate(self.imageView.transform,deltaX,deltaY);
        transform = CGAffineTransformRotate(transform, recognizer.rotation);
        transform = CGAffineTransformTranslate(transform, -deltaX, -deltaY);
        self.imageView.transform = transform;

        recognizer.rotation = 0;
    }

}

- (IBAction)handlePinch:(UIPinchGestureRecognizer *)recognizer
{
    if([self handleGestureState:recognizer.state]) {
        if(recognizer.state == UIGestureRecognizerStateBegan){
            self.scaleCenter = self.touchCenter;
        }
        CGFloat deltaX = self.scaleCenter.x-self.imageView.bounds.size.width/2.0;
        CGFloat deltaY = self.scaleCenter.y-self.imageView.bounds.size.height/2.0;
        
        if(self.scale*recognizer.scale > kMinZoomScale
           && self.scale*recognizer.scale < kMaxZoomScale){
            CGAffineTransform transform =  CGAffineTransformTranslate(self.imageView.transform, deltaX, deltaY);
            transform = CGAffineTransformScale(transform, recognizer.scale, recognizer.scale);
            transform = CGAffineTransformTranslate(transform, -deltaX, -deltaY);
            self.imageView.transform = transform;
        }
        
        self.scale *= recognizer.scale;
        recognizer.scale = 1;
    }
}

- (IBAction)handleTap:(UITapGestureRecognizer *)recogniser {
    [self reset:YES];
}

- (BOOL)gestureRecognizer:(UIGestureRecognizer *)gestureRecognizer shouldRecognizeSimultaneouslyWithGestureRecognizer:(UIGestureRecognizer *)otherGestureRecognizer {
    return YES;
}

# pragma mark Image Transformation


- (UIImage *)scaledImage:(UIImage *)source toSize:(CGSize)size withQuality:(CGInterpolationQuality)quality
{
    CGImageRef cgImage  = [self newScaledImage:source.CGImage withOrientation:source.imageOrientation toSize:size withQuality:quality];
    UIImage * result = [UIImage imageWithCGImage:cgImage scale:1.0 orientation:UIImageOrientationUp];
    CGImageRelease(cgImage);
    return result;
}


- (CGImageRef)newScaledImage:(CGImageRef)source withOrientation:(UIImageOrientation)orientation toSize:(CGSize)size withQuality:(CGInterpolationQuality)quality
{
    CGSize srcSize = size;
    CGFloat rotation = 0.0;
    
    switch(orientation)
    {
        case UIImageOrientationUp: {
            rotation = 0;
        } break;
        case UIImageOrientationDown: {
            rotation = M_PI;
        } break;
        case UIImageOrientationLeft:{
            rotation = M_PI_2;
            srcSize = CGSizeMake(size.height, size.width);
        } break;
        case UIImageOrientationRight: {
            rotation = -M_PI_2;
            srcSize = CGSizeMake(size.height, size.width);
        } break;
        default:
            break;
    }
    
    CGContextRef context = CGBitmapContextCreate(NULL,
                                                 size.width,
                                                 size.height,
                                                 8, //CGImageGetBitsPerComponent(source),
                                                 0,
                                                 CGImageGetColorSpace(source),
                                                 kCGImageAlphaNoneSkipFirst//CGImageGetBitmapInfo(source)
                                                 );
    
    CGContextSetInterpolationQuality(context, quality);
    CGContextTranslateCTM(context,  size.width/2,  size.height/2);
    CGContextRotateCTM(context,rotation);
    
    CGContextDrawImage(context, CGRectMake(-srcSize.width/2 ,
                                           -srcSize.height/2,
                                           srcSize.width,
                                           srcSize.height),
                       source);
    
    CGImageRef resultRef = CGBitmapContextCreateImage(context);
    CGContextRelease(context);

    return resultRef;
}

- (CGImageRef)newTransformedImage:(CGAffineTransform)transform
                     sourceImage:(CGImageRef)sourceImage
                    sourceSize:(CGSize)sourceSize
           sourceOrientation:(UIImageOrientation)sourceOrientation
                 outputWidth:(CGFloat)outputWidth
                    cropSize:(CGSize)cropSize
               imageViewSize:(CGSize)imageViewSize
{
    CGImageRef source = [self newScaledImage:sourceImage
                         withOrientation:sourceOrientation
                                  toSize:sourceSize
                             withQuality:kCGInterpolationNone];
    
    CGFloat aspect = cropSize.height/cropSize.width;
    CGSize outputSize = CGSizeMake(outputWidth, outputWidth*aspect);
    
    CGContextRef context = CGBitmapContextCreate(NULL,
                                                 outputSize.width,
                                                 outputSize.height,
                                                 CGImageGetBitsPerComponent(source),
                                                 0,
                                                 CGImageGetColorSpace(source),
                                                 CGImageGetBitmapInfo(source));
    
    CGContextSetFillColorWithColor(context,  [[UIColor clearColor] CGColor]);
    CGContextFillRect(context, CGRectMake(0, 0, outputSize.width, outputSize.height));
    
    CGAffineTransform uiCoords = CGAffineTransformMakeScale(outputSize.width/cropSize.width,
                                                            outputSize.height/cropSize.height);
    uiCoords = CGAffineTransformTranslate(uiCoords, cropSize.width/2.0, cropSize.height/2.0);
    uiCoords = CGAffineTransformScale(uiCoords, 1.0, -1.0);
    CGContextConcatCTM(context, uiCoords);
    
    CGContextConcatCTM(context, transform);
    CGContextScaleCTM(context, 1.0, -1.0);
   
    CGContextDrawImage(context,
                       CGRectMake(-imageViewSize.width/2.0,
                                           -imageViewSize.height/2.0,
                                           imageViewSize.width,
                                           imageViewSize.height)
                       ,source);
    
    
    CGImageRef resultRef = CGBitmapContextCreateImage(context);
    CGContextRelease(context);
    CGImageRelease(source);
    return resultRef;
}


- (CGRect)cropBoundsInSourceImage
{
    CGAffineTransform uiCoords = CGAffineTransformMakeScale(self.sourceImage.size.width/self.imageView.bounds.size.width,
                                                            self.sourceImage.size.height/self.imageView.bounds.size.height);
    
    uiCoords = CGAffineTransformTranslate(uiCoords, self.imageView.bounds.size.width/2.0, self.imageView.bounds.size.height/2.0);
    uiCoords = CGAffineTransformScale(uiCoords, 1.0, -1.0);

    CGRect crop =  CGRectMake(-self.cropSize.width/2.0, -self.cropSize.height/2.0, self.cropSize.width, self.cropSize.height);
    return CGRectApplyAffineTransform(crop, CGAffineTransformConcat(CGAffineTransformInvert(self.imageView.transform),uiCoords));
}
#pragma mark Subclass Hooks

- (void)startTransformHook
{
}

- (void)endTransformHook
{
}



@end
